"""Model selection extension for PEtab."""

from .candidate_space import *
from .constants import *
from .criteria import *
from .misc import *
from .model import *
from .model_space import *
from .model_subspace import *
from .problem import *
from .ui import *



"""Constants for the PEtab Select package."""
from enum import Enum
from pathlib import Path
from typing import Dict, List, Union

# Zero-indexed column/row indices
MODEL_ID_COLUMN = 0
PETAB_YAML_COLUMN = 1
# It is assumed that all columns after PARAMETER_DEFINITIONS_START contain
# parameter IDs.
PARAMETER_DEFINITIONS_START = 2
HEADER_ROW = 0

PARAMETER_VALUE_DELIMITER = ';'
CODE_DELIMITER = '-'
ESTIMATE = 'estimate'
PETAB_ESTIMATE_FALSE = 0
PETAB_ESTIMATE_TRUE = 1

# TYPING_PATH = Union[str, Path]
TYPE_PATH = Union[str, Path]

# Model space file columns
# TODO ensure none of these occur twice in the column header (this would
#      suggest that a parameter has a conflicting name)
# MODEL_ID = 'modelId'  # TODO already defined, reorganize constants
# YAML = 'YAML'  # FIXME
MODEL_ID = 'model_id'
MODEL_SUBSPACE_ID = 'model_subspace_id'
MODEL_SUBSPACE_INDICES = 'model_subspace_indices'
MODEL_CODE = 'model_code'
MODEL_HASH = 'model_hash'
MODEL_HASHES = 'model_hashes'
# If `predecessor_model_hash` is defined for a model, it is the ID of the model that the
# current model was/is to be compared to. This is part of the result and is
# only (optionally) set by the PEtab calibration tool. It is not defined by the
# PEtab Select model selection problem (but may be subsequently stored in the
# PEtab Select model report format.
PREDECESSOR_MODEL_HASH = 'predecessor_model_hash'
PETAB_PROBLEM = 'petab_problem'
PETAB_YAML = 'petab_yaml'
SBML = 'sbml'
HASH = 'hash'

# MODEL_SPACE_FILE_NON_PARAMETER_COLUMNS = [MODEL_ID, PETAB_YAML]
MODEL_SPACE_FILE_NON_PARAMETER_COLUMNS = [MODEL_SUBSPACE_ID, PETAB_YAML]

# COMPARED_MODEL_ID = 'compared_'+MODEL_ID
YAML_FILENAME = 'yaml'

# FORWARD = 'forward'
# BACKWARD = 'backward'
# BIDIRECTIONAL = 'bidirectional'
# LATERAL = 'lateral'


# DISTANCES = {
#    FORWARD: {
#        'l1': 1,
#        'size': 1,
#    },
#    BACKWARD: {
#        'l1': 1,
#        'size': -1,
#    },
#    LATERAL: {
#        'l1': 2,
#        'size': 0,
#    },
# }

CRITERIA = 'criteria'
# FIXME remove, change all uses to Enum below
# AIC = 'AIC'
# AICC = 'AICc'
# BIC = 'BIC'
# AKAIKE_INFORMATION_CRITERION = AIC
# CORRECTED_AKAIKE_INFORMATION_CRITERION = AICC
# BAYESIAN_INFORMATION_CRITERION = BIC
# LH = 'LH'
# LLH = 'LLH'
# NLLH = 'NLLH'
# LIKELIHOOD = LH
# LOG_LIKELIHOOD = LLH
# NEGATIVE_LOG_LIKELIHOOD = NLLH


PARAMETERS = 'parameters'
# PARAMETER_ESTIMATE = 'parameter_estimate'
ESTIMATED_PARAMETERS = 'estimated_parameters'

CRITERION = 'criterion'
METHOD = 'method'
VERSION = 'version'
MODEL_SPACE_FILES = 'model_space_files'

CANDIDATE_SPACE_ARGUMENTS = 'candidate_space_arguments'
METHOD_SCHEME = 'method_scheme'
PREVIOUS_METHODS = 'previous_methods'
NEXT_METHOD = 'next_method'
PREDECESSOR_MODEL = 'predecessor_model'

MODEL = 'model'
MODELS = 'models'

# Parameters can be fixed to a value, or estimated if indicated with the string
# `ESTIMATE`.
# TODO change to `Literal[ESTIMATE]` (Python >= 3.8)
TYPE_PARAMETER = Union[float, int, ESTIMATE]
TYPE_PARAMETER_OPTIONS = List[TYPE_PARAMETER]
# Parameter ID -> parameter value mapping.
TYPE_PARAMETER_DICT = Dict[str, TYPE_PARAMETER]
# Parameter ID -> multiple possible parameter values.
TYPE_PARAMETER_OPTIONS_DICT = Dict[str, TYPE_PARAMETER_OPTIONS]

TYPE_CRITERION = float


class Method(str, Enum):
    """String literals for model selection methods."""

    BACKWARD = 'backward'
    BIDIRECTIONAL = 'bidirectional'
    BRUTE_FORCE = 'brute_force'
    FAMOS = 'famos'
    FORWARD = 'forward'
    FORWARD_AND_BACKWARD = 'forward_and_backward'
    LATERAL = 'lateral'
    MOST_DISTANT = 'most_distant'


class Criterion(str, Enum):
    """String literals for model selection criteria."""

    AIC = 'AIC'
    AICC = 'AICc'
    BIC = 'BIC'
    LH = 'LH'
    LLH = 'LLH'
    NLLH = 'NLLH'


# Methods that move through model space by taking steps away from some model.
STEPWISE_METHODS = [
    Method.BACKWARD,
    Method.BIDIRECTIONAL,
    Method.FORWARD,
    Method.FORWARD_AND_BACKWARD,
    Method.LATERAL,
]
# Methods that require an initial model.
INITIAL_MODEL_METHODS = [
    Method.BACKWARD,
    Method.BIDIRECTIONAL,
    Method.FORWARD,
    Method.FORWARD_AND_BACKWARD,
    Method.LATERAL,
]

# Virtual initial models can be used to initialize some initial model methods.
VIRTUAL_INITIAL_MODEL = 'virtual_initial_model'
VIRTUAL_INITIAL_MODEL_METHODS = [
    Method.BACKWARD,
    Method.BIDIRECTIONAL,
    Method.FORWARD,
    Method.FORWARD_AND_BACKWARD,
]



"""Implementations of model selection criteria."""

import math

import petab
from petab.C import OBJECTIVE, OBJECTIVE_PRIOR_PARAMETERS, OBJECTIVE_PRIOR_TYPE

from .constants import (  # LH,; LLH,; NLLH,
    PETAB_PROBLEM,
    TYPE_CRITERION,
    Criterion,
)

# from .model import Model


# use as attribute e.g. `Model.criterion_computer`?
class CriterionComputer:
    """Compute various criterion."""

    def __init__(
        self,
        model: 'petab_select.Model',
    ):
        self.model = model
        self._petab_problem = None

    @property
    def petab_problem(self) -> petab.Problem:
        """The PEtab problem that corresponds to the model.

        Implemented as a property such that the `petab.Problem` object
        is only constructed if explicitly requested.

        Improves speed of operations on models by a lot. For example, analysis of models
        that already have criteria computed can skip loading their PEtab problem again.
        """
        # TODO refactor, if `petab_problem` is going to be produced here anyway, store
        #      in model instance instead, for use elsewhere (e.g. pyPESTO)
        #      i.e.: this is a property of a `Model` instance, not `CriterionComputer`
        if self._petab_problem is None:
            self._petab_problem = self.model.to_petab()[PETAB_PROBLEM]
        return self._petab_problem

    def __call__(self, criterion: Criterion) -> float:
        """Get a criterion value.

        Args:
            criterion:
                The ID of the criterion.

        Returns:
            The criterion value.
        """
        return getattr(self, 'get_' + criterion.value.lower())()

    def get_aic(self) -> float:
        """Get the Akaike information criterion."""
        return calculate_aic(
            nllh=self.get_nllh(),
            n_estimated=self.get_n_estimated(),
        )

    def get_aicc(self) -> float:
        """Get the corrected Akaike information criterion."""
        return calculate_aicc(
            nllh=self.get_nllh(),
            n_estimated=self.get_n_estimated(),
            n_measurements=self.get_n_measurements(),
            n_priors=self.get_n_priors(),
        )

    def get_bic(self) -> float:
        """Get the Bayesian information criterion."""
        return calculate_bic(
            nllh=self.get_nllh(),
            n_estimated=self.get_n_estimated(),
            n_measurements=self.get_n_measurements(),
            n_priors=self.get_n_priors(),
        )

    def get_nllh(self) -> float:
        """Get the negative log-likelihood."""
        nllh = self.model.get_criterion(Criterion.NLLH, compute=False)
        if nllh is None:
            nllh = -1 * self.get_llh()
        return nllh

    def get_llh(self) -> float:
        """Get the log-likelihood."""
        llh = self.model.get_criterion(Criterion.LLH, compute=False)
        if llh is None:
            llh = math.log(self.get_lh())
        return llh

    def get_lh(self) -> float:
        """Get the likelihood."""
        lh = self.model.get_criterion(Criterion.LH, compute=False)
        llh = self.model.get_criterion(Criterion.LLH, compute=False)
        nllh = self.model.get_criterion(Criterion.NLLH, compute=False)

        if lh is not None:
            return lh
        elif llh is not None:
            return math.exp(llh)
        elif nllh is not None:
            return math.exp(-1 * nllh)

        raise ValueError(
            'Please supply the likelihood (LH) or a compatible transformation. Compatible transformations: log(LH), -log(LH).'
        )

    def get_n_estimated(self) -> int:
        """Get the number of estimated parameters."""
        return len(self.petab_problem.x_free_indices)

    def get_n_measurements(self) -> int:
        """Get the number of measurements."""
        return len(self.petab_problem.measurement_df)

    def get_n_priors(self) -> int:
        """Get the number of priors."""
        df = self.petab_problem.parameter_df

        # At least one of the objective prior columns should be present.
        if not (
            OBJECTIVE_PRIOR_TYPE in df or OBJECTIVE_PRIOR_PARAMETERS in df
        ):
            return 0

        # If both objective prior columns are not present, raise an error.
        if not (
            OBJECTIVE_PRIOR_TYPE in df and OBJECTIVE_PRIOR_PARAMETERS in df
        ):
            raise NotImplementedError(
                'Currently expect that prior types are specified with prior parameters (no default values). Please provide an example for implementation.'
            )

        # Expect that the number of non-empty values in both objective prior columns
        # are the same.
        if not (
            df[OBJECTIVE_PRIOR_TYPE].notna().sum()
            == df[OBJECTIVE_PRIOR_PARAMETERS].notna().sum()
        ):
            raise NotImplementedError(
                'Some objective prior values are missing.'
            )

        number_of_priors = df[OBJECTIVE_PRIOR_TYPE].notna().sum()
        return number_of_priors

    # def get_criterion(self, id: str) -> TYPE_CRITERION:
    #    """Get a criterion value, by criterion ID.
    #    FIXME: superseded by `__call__`

    #    id:
    #        The ID of the criterion (e.g. `petab_select.constants.Criterion.AIC`).

    #    Returns:
    #        The criterion value.
    #    """
    #    return getattr(self, f'get_{id}')()


# TODO should fixed parameters count as measurements/priors when comparing to models
#      that estimate the same parameters?
def calculate_aic(
    nllh: float,
    n_estimated: int,
) -> float:
    """
    Calculate the Akaike information criterion (AIC) for a model.

    Args:
        nllh:
            The negative log likelihood.
        n_estimated:
            The number of estimated parameters in the model.

    Returns:
        The AIC value.
    """
    return 2 * (n_estimated + nllh)


def calculate_aicc(
    nllh: float,
    n_estimated: int,
    n_measurements: int,
    n_priors: int,
) -> float:
    """
    Calculate the corrected Akaike information criterion (AICc) for a model.

    Args:
        nllh:
            The negative log likelihood.
        n_estimated:
            The number of estimated parameters in the model.
        n_measurements:
            The number of measurements used in the likelihood.
        n_priors:
            The number of priors used in the objective function.

    Returns:
        The AICc value.
    """
    return calculate_aic(n_estimated, nllh) + 2 * n_estimated * (
        n_estimated + 1
    ) / (n_measurements + n_priors - n_estimated - 1)


def calculate_bic(
    nllh: float,
    n_estimated: int,
    n_measurements: int,
    n_priors: int,
):
    """
    Calculate the Bayesian information criterion (BIC) for a model.

    Args
        nllh:
            The negative log likelihood.
        n_estimated:
            The number of estimated parameters in the model.
        n_measurements:
            The number of measurements used in the likelihood.
        n_priors:
            The number of priors used in the objective function.

    Returns:
        The BIC value.
    """
    return n_estimated * math.log(n_measurements + n_priors) + 2 * nllh



"""The `Model` class."""
import abc
import warnings
from os.path import relpath
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import petab
import yaml
from more_itertools import one
from petab.C import ESTIMATE, NOMINAL_VALUE

from .constants import (
    CRITERIA,
    ESTIMATED_PARAMETERS,
    MODEL_HASH,
    MODEL_ID,
    MODEL_SUBSPACE_ID,
    MODEL_SUBSPACE_INDICES,
    PARAMETERS,
    PETAB_ESTIMATE_TRUE,
    PETAB_PROBLEM,
    PETAB_YAML,
    PREDECESSOR_MODEL_HASH,
    TYPE_CRITERION,
    TYPE_PARAMETER,
    TYPE_PATH,
    VIRTUAL_INITIAL_MODEL,
    Criterion,
)
from .criteria import CriterionComputer
from .misc import (
    hash_list,
    hash_parameter_dict,
    hash_str,
    parameter_string_to_value,
)
from .petab import PetabMixin


class Model(PetabMixin):
    """A (possibly uncalibrated) model.

    NB: some of these attribute names correspond to constants defined in the
    `constants.py` file, to facilitate loading models from/saving models to
    disk (see the `saved_attributes` attribute).

    Attributes:
        converters_load:
            Functions to convert attributes from YAML to `Model`.
        converters_save:
            Functions to convert attributes from `Model` to YAML.
        criteria:
            The criteria values of the calibrated model (e.g. AIC).
        hash_attributes:
            This attribute is currently not used.
            Attributes that will be used to calculate the hash of the
            `Model` instance. NB: this hash is used during pairwise comparison
            to determine whether any two `Model` instances are unique. The
            model instances are compared by their parameter estimation
            problems, as opposed to parameter estimation results, which may
            differ due to e.g. floating-point arithmetic.
        model_id:
            The model ID.
        petab_yaml:
            The path to the PEtab problem YAML file.
        parameters:
            Parameter values that will overwrite the PEtab problem definition,
            or change parameters to be estimated.
        estimated_parameters:
            Parameter estimates from a model calibration tool, for parameters
            that are specified as estimated in the PEtab problem or PEtab
            Select model YAML.
        saved_attributes:
            Attributes that will be saved to disk by the `Model.to_yaml`
            method.
    """

    saved_attributes = (
        MODEL_ID,
        MODEL_SUBSPACE_ID,
        MODEL_SUBSPACE_INDICES,
        MODEL_HASH,
        PREDECESSOR_MODEL_HASH,
        PETAB_YAML,
        PARAMETERS,
        ESTIMATED_PARAMETERS,
        CRITERIA,
    )
    converters_load = {
        MODEL_ID: lambda x: x,
        MODEL_SUBSPACE_ID: lambda x: x,
        MODEL_SUBSPACE_INDICES: lambda x: [] if not x else x,
        MODEL_HASH: lambda x: x,
        PREDECESSOR_MODEL_HASH: lambda x: x,
        PETAB_YAML: lambda x: x,
        PARAMETERS: lambda x: x,
        ESTIMATED_PARAMETERS: lambda x: x,
        CRITERIA: lambda x: {
            # `criterion_id_value` is the ID of the criterion in the enum `Criterion`.
            Criterion(criterion_id_value): float(criterion_value)
            for criterion_id_value, criterion_value in x.items()
        },
    }
    converters_save = {
        MODEL_ID: lambda x: str(x),
        MODEL_SUBSPACE_ID: lambda x: str(x),
        MODEL_SUBSPACE_INDICES: lambda x: [int(xi) for xi in x],
        MODEL_HASH: lambda x: str(x),
        PREDECESSOR_MODEL_HASH: lambda x: str(x),
        PETAB_YAML: lambda x: str(x),
        PARAMETERS: lambda x: {str(k): v for k, v in x.items()},
        # FIXME handle with a `set_estimated_parameters` method instead?
        # to avoid `float` cast here. Reason for cast is because e.g. pyPESTO
        # can provide type `np.float64`, which causes issues when writing to
        # YAML.
        # ESTIMATED_PARAMETERS: lambda x: x,
        ESTIMATED_PARAMETERS: lambda x: {
            str(id): float(value) for id, value in x.items()
        },
        CRITERIA: lambda x: {
            criterion_id.value: float(criterion_value)
            for criterion_id, criterion_value in x.items()
        },
    }
    hash_attributes = {
        # MODEL_ID: lambda x: hash(x),  # possible circular dependency on hash
        # MODEL_SUBSPACE_ID: lambda x: hash(x),
        # MODEL_SUBSPACE_INDICES: hash_list,
        # TODO replace `YAML` with `PETAB_PROBLEM_HASH`, as YAML could refer to
        #      different problems if used on different filesystems or sometimes
        #      absolute and other times relative. Better to check whether the
        #      PEtab problem itself is unique.
        # TODO replace `PARAMETERS` with `PARAMETERS_ALL`, which should be al
        #      parameters in the PEtab problem. This avoids treating the PEtab problem
        #      differently to the model (in a subspace with the PEtab problem) that has
        #      all nominal values defined in the subspace.
        # TODO add `estimated_parameters`? Needs to be clarified whether this hash
        #      should be unique amongst only not-yet-calibrated models, or may also
        #      return the same value between differently parameterized models that ended
        #      up being calibrated to be the same... probably should be the former.
        #      Currently, the hash is stored, hence will "persist" after calibration
        #      if the same `Model` instance is used.
        # PETAB_YAML: lambda x: hash(x),
        PETAB_YAML: hash_str,
        PARAMETERS: hash_parameter_dict,
    }

    def __init__(
        self,
        petab_yaml: TYPE_PATH,
        model_subspace_id: str = None,
        model_id: str = None,
        model_subspace_indices: List[int] = None,
        predecessor_model_hash: str = None,
        parameters: Dict[str, Union[int, float]] = None,
        estimated_parameters: Dict[str, Union[int, float]] = None,
        criteria: Dict[str, float] = None,
        # Optionally provided to reduce repeated parsing of `petab_yaml`.
        petab_problem: Optional[petab.Problem] = None,
        model_hash: Optional[Any] = None,
    ):
        self.model_id = model_id
        self.model_subspace_id = model_subspace_id
        self.model_subspace_indices = model_subspace_indices
        # TODO clean parameters, ensure single float or str (`ESTIMATE`) type
        self.parameters = parameters
        self.estimated_parameters = estimated_parameters
        self.criteria = criteria
        self.model_hash = model_hash

        self.predecessor_model_hash = predecessor_model_hash

        if self.parameters is None:
            self.parameters = {}
        if self.estimated_parameters is None:
            self.estimated_parameters = {}
        if self.criteria is None:
            self.criteria = {}

        super().__init__(petab_yaml=petab_yaml, petab_problem=petab_problem)

        if self.model_id is None:
            self.model_id = self.get_hash()

        self.criterion_computer = CriterionComputer(self)

    def set_criterion(self, criterion: Criterion, value: float) -> None:
        """Set a criterion value for the model.

        Args:
            criterion:
                The criterion (e.g. `petab_select.constants.Criterion.AIC`).
            value:
                The criterion value for the (presumably calibrated) model.
        """
        if criterion in self.criteria:
            warnings.warn(
                'Overwriting saved criterion value. '
                f'Criterion: {criterion}. Value: {self.get_criterion(criterion)}.'
            )
            # FIXME debug why value is overwritten during test case 0002.
            if False:
                print(
                    'Overwriting saved criterion value. '
                    f'Criterion: {criterion}. Value: {self.get_criterion(criterion)}.'
                )
                breakpoint()
        self.criteria[criterion] = value

    def has_criterion(self, criterion: Criterion) -> bool:
        """Check whether the model provides a value for a criterion.

        Args:
            criterion:
                The criterion (e.g. `petab_select.constants.Criterion.AIC`).
        """
        # TODO also `and self.criteria[id] is not None`?
        return criterion in self.criteria

    def get_criterion(
        self,
        criterion: Criterion,
        compute: bool = True,
    ) -> Union[TYPE_CRITERION, None]:
        """Get a criterion value for the model.

        Args:
            criterion:
                The ID of the criterion (e.g. `petab_select.constants.Criterion.AIC`).
            compute:
                Whether to try to compute the criterion value based on other model
                attributes. For example, if the `'AIC'` criterion is requested, this
                can be computed from a predetermined model likelihood and its
                number of estimated parameters.

        Returns:
            The criterion value, or `None` if it is not available.
            TODO check for previous use of this method before `.get` was used
        """
        if criterion not in self.criteria and compute:
            self.compute_criterion(criterion=criterion)
            # value = self.criterion_computer(criterion=id)
            # self.set_criterion(id=id, value=value)

        return self.criteria.get(criterion, None)

    def compute_criterion(self, criterion: Criterion) -> TYPE_CRITERION:
        """Compute a criterion value for the model.

        The value will also be stored, which will overwrite any previously stored value
        for the criterion.

        Args:
            id:
                The ID of the criterion (e.g. `petab_select.constants.Criterion.AIC`).

        Returns:
            The criterion value.
        """
        criterion_value = self.criterion_computer(criterion)
        self.set_criterion(criterion, criterion_value)
        return criterion_value

    @staticmethod
    def from_dict(
        model_dict: Dict[str, Any],
        base_path: TYPE_PATH = None,
        petab_problem: petab.Problem = None,
    ) -> 'Model':
        """Generate a model from a dictionary of attributes.

        Args:
            model_dict:
                A dictionary of attributes. The keys are attribute
                names, the values are the corresponding attribute values for
                the model. Required attributes are the required arguments of
                the `Model.__init__` method.
            base_path:
                The path that any relative paths in the model are relative to
                (e.g. the path to the PEtab problem YAML file
                `Model.petab_yaml` may be relative).
            petab_problem:
                Optionally provide the PEtab problem, to avoid loading it multiple
                times.
                NB: This may causes issues if multiple models write to the same PEtab
                problem in memory.

        Returns:
            A model instance, initialized with the provided attributes.
        """
        unknown_attributes = set(model_dict).difference(Model.converters_load)
        if unknown_attributes:
            warnings.warn(
                'Ignoring unknown attributes: ' + ', '.join(unknown_attributes)
            )

        if base_path is not None:
            model_dict[PETAB_YAML] = base_path / model_dict[PETAB_YAML]

        model_dict = {
            attribute: Model.converters_load[attribute](value)
            for attribute, value in model_dict.items()
            if attribute in Model.converters_load
        }
        model_dict[PETAB_PROBLEM] = petab_problem
        return Model(**model_dict)

    @staticmethod
    def from_yaml(model_yaml: TYPE_PATH) -> 'Model':
        """Generate a model from a PEtab Select model YAML file.

        Args:
            model_yaml:
                The path to the PEtab Select model YAML file.

        Returns:
            A model instance, initialized with the provided attributes.
        """
        with open(str(model_yaml)) as f:
            model_dict = yaml.safe_load(f)
        # TODO check that the hash is reproducible
        if isinstance(model_dict, list):
            try:
                model_dict = one(model_dict)
            except ValueError:
                if len(model_dict) <= 1:
                    raise
                raise ValueError(
                    'The provided YAML file contains a list with greater than '
                    'one element. Use the `models_from_yaml_list` method or '
                    'provide a PEtab Select model YAML file with only one '
                    'model specified.'
                )

        return Model.from_dict(model_dict, base_path=Path(model_yaml).parent)

    def to_dict(
        self,
        resolve_paths: bool = True,
        paths_relative_to: Union[str, Path] = None,
    ) -> Dict[str, Any]:
        """Generate a dictionary from the attributes of a `Model` instance.

        Args:
            resolve_paths:
                Whether to resolve relative paths into absolute paths.
            paths_relative_to:
                If not `None`, paths will be converted to be relative to this path.
                Takes priority over `resolve_paths`.

        Returns:
            A dictionary of attributes. The keys are attribute
            names, the values are the corresponding attribute values for
            the model. Required attributes are the required arguments of
            the `Model.__init__` method.
        """
        model_dict = {}
        for attribute in self.saved_attributes:
            model_dict[attribute] = self.converters_save[attribute](
                getattr(self, attribute)
            )
        # TODO test
        if resolve_paths:
            if model_dict[PETAB_YAML]:
                model_dict[PETAB_YAML] = str(
                    Path(model_dict[PETAB_YAML]).resolve()
                )
        if paths_relative_to is not None:
            if model_dict[PETAB_YAML]:
                model_dict[PETAB_YAML] = relpath(
                    Path(model_dict[PETAB_YAML]).resolve(),
                    Path(paths_relative_to).resolve(),
                )
        return model_dict

    def to_yaml(self, petab_yaml: TYPE_PATH, *args, **kwargs) -> None:
        """Generate a PEtab Select model YAML file from a `Model` instance.

        Parameters:
            petab_yaml:
                The location where the PEtab Select model YAML file will be
                saved.
            args, kwargs:
                Additional arguments are passed to `self.to_dict`.
        """
        # FIXME change `getattr(self, PETAB_YAML)` to be relative to
        # destination?
        # kind of fixed, as the path will be resolved in `to_dict`.
        with open(petab_yaml, 'w') as f:
            yaml.dump(self.to_dict(*args, **kwargs), f)
        # yaml.dump(self.to_dict(), str(petab_yaml))

    def to_petab(
        self,
        output_path: TYPE_PATH = None,
    ) -> Tuple[petab.Problem, TYPE_PATH]:
        """Generate a PEtab problem.

        Args:
            output_path:
                The directory where PEtab files will be written to disk. If not
                specified, the PEtab files will not be written to disk.

        Returns:
            A 2-tuple. The first value is a PEtab problem that can be used
            with a PEtab-compatible tool for calibration of this model. If
            `output_path` is not `None`, the second value is the path to a
            PEtab YAML file that can be used to load the PEtab problem (the
            first value) into any PEtab-compatible tool. If
        """
        # TODO could use `copy.deepcopy(self.petab_problem)` from PetabMixin?
        petab_problem = petab.Problem.from_yaml(str(self.petab_yaml))
        for parameter_id, parameter_value in self.parameters.items():
            # If the parameter is to be estimated.
            if parameter_value == ESTIMATE:
                petab_problem.parameter_df.loc[parameter_id, ESTIMATE] = 1
            # Else the parameter is to be fixed.
            else:
                petab_problem.parameter_df.loc[parameter_id, ESTIMATE] = 0
                petab_problem.parameter_df.loc[
                    parameter_id, NOMINAL_VALUE
                ] = parameter_string_to_value(parameter_value)
                # parameter_value

        petab_yaml = None
        if output_path is not None:
            output_path = Path(output_path)
            output_path.mkdir(exist_ok=True, parents=True)
            petab_yaml = petab_problem.to_files_generic(
                prefix_path=output_path
            )

        return {
            PETAB_PROBLEM: petab_problem,
            PETAB_YAML: petab_yaml,
        }

    def get_hash(self) -> int:
        """Get the model hash.

        Currently designed to only use pre-calibration information, such that if a model
        is calibrated twice and the two calibrated models differ in their parameter
        estimates, then they will still have the same hash.

        This is not implemented as `__hash__` because Python automatically truncates
        values in a system-dependent manner, which reduces interoperability
        ( https://docs.python.org/3/reference/datamodel.html#object.__hash__ ).

        Returns:
            The hash.
        """
        if self.model_hash is None:
            self.model_hash = hash_list(
                [
                    method(getattr(self, attribute))
                    for attribute, method in Model.hash_attributes.items()
                ]
            )
        return self.model_hash

    def __hash__(self) -> None:
        """Use `Model.get_hash` instead."""
        raise NotImplementedError('Use `Model.get_hash() instead.`')

    def __str__(self):
        """Get a print-ready string representation of the model.

        Returns:
            The print-ready string representation, in TSV format.
        """
        parameter_ids = '\t'.join(self.parameters.keys())
        parameter_values = '\t'.join(str(v) for v in self.parameters.values())
        header = '\t'.join([MODEL_ID, PETAB_YAML, parameter_ids])
        data = '\t'.join(
            [self.model_id, str(self.petab_yaml), parameter_values]
        )
        # header = f'{MODEL_ID}\t{PETAB_YAML}\t{parameter_ids}'
        # data = f'{self.model_id}\t{self.petab_yaml}\t{parameter_values}'
        return f'{header}\n{data}'

    def get_mle(self) -> Dict[str, float]:
        """Get the maximum likelihood estimate of the model.

        # Check if original PEtab problem or PEtab Select model has estimated
        # parameters. e.g. can use some of `self.to_petab` to get the parameter
        # df and see if any are estimated.
        if not self.has_estimated_parameters:
            warn('The MLE for this model contains no estimated parameters.')
        if not all([
            parameter_id in getattr(self, ESTIMATED_PARAMETERS)
            for parameter_id in self.get_estimated_parameter_ids_all()
        ]):
            warn('Not all estimated parameters have estimates stored.')
        petab_problem = petab.Problem.from_yaml(str(self.petab_yaml))
        return {
            parameter_id: (
                getattr(self, ESTIMATED_PARAMETERS).get(
                    # Return estimated parameter from `petab_select.Model`
                    # if possible.
                    parameter_id,
                    # Else return nominal value from PEtab parameter table.
                    petab_problem.parameter_df.loc[
                        parameter_id, NOMINAL_VALUE
                    ],
                )
            )
            for parameter_id in petab_problem.parameter_df.index
        }
        # TODO rewrite to construct return dict in a for loop, for more
        # informative error message as soon as a "should-be-estimated"
        # parameter has not estimate available in `self.estimated_parameters`.
        """
        # TODO
        pass

    def get_estimated_parameter_ids_all(self) -> List[str]:
        estimated_parameter_ids = []

        # Add all estimated parameters in the PEtab problem.
        petab_problem = petab.Problem.from_yaml(str(self.petab_yaml))
        for parameter_id in petab_problem.parameter_df.index:
            if (
                petab_problem.parameter_df.loc[parameter_id, ESTIMATE]
                == PETAB_ESTIMATE_TRUE
            ):
                estimated_parameter_ids.append(parameter_id)

        # Add additional estimated parameters, and collect fixed parameters,
        # in this model's parameterization.
        fixed_parameter_ids = []
        for parameter_id, value in self.parameters.items():
            if (
                value == ESTIMATE
                and parameter_id not in estimated_parameter_ids
            ):
                estimated_parameter_ids.append(parameter_id)
            elif value != ESTIMATE:
                fixed_parameter_ids.append(parameter_id)

        # Remove fixed parameters.
        estimated_parameter_ids = [
            parameter_id
            for parameter_id in estimated_parameter_ids
            if parameter_id not in fixed_parameter_ids
        ]

        return estimated_parameter_ids

    def get_parameter_values(
        self,
        parameter_ids: Optional[List[str]] = None,
    ) -> List[TYPE_PARAMETER]:
        """Get parameter values.

        Includes `ESTIMATE` for parameters that should be estimated.

        The ordering is by `parameter_ids` if supplied, else
        `self.petab_parameters`.

        Args:
            parameter_ids:
                The IDs of parameters that values will be returned for. Order
                is maintained.

        Returns:
            The values of parameters.
        """
        if parameter_ids is None:
            parameter_ids = list(self.petab_parameters)
        return [
            self.parameters.get(
                parameter_id,
                # Default to PEtab problem.
                self.petab_parameters[parameter_id],
            )
            for parameter_id in parameter_ids
        ]


def default_compare(
    model0: Model,
    model1: Model,
    criterion: Criterion,
    criterion_threshold: float = 0,
) -> bool:
    """Compare two calibrated models by their criterion values.

    It is assumed that the model `model0` provides a value for the criterion
    `criterion`, or is the `VIRTUAL_INITIAL_MODEL`.

    Args:
        model0:
            The original model.
        model1:
            The new model.
        criterion:
            The criterion by which models will be compared.
        criterion_threshold:
            The value by which the new model must improve on the original
            model. Should be non-negative.

    Returns:
        `True` if `model1` has a better criterion value than `model0`, else
        `False`.
    """
    if not model1.has_criterion(criterion):
        warnings.warn(
            f'Model "{model1.model_id}" does not provide a value for the criterion "{criterion}".'
        )
        return False
    if model0 == VIRTUAL_INITIAL_MODEL or model0 is None:
        return True
    if criterion_threshold < 0:
        warnings.warn(
            'The provided criterion threshold is negative. The absolute value will be used instead.'
        )
        criterion_threshold = abs(criterion_threshold)
    if criterion in [
        Criterion.AIC,
        Criterion.AICC,
        Criterion.BIC,
        Criterion.NLLH,
    ]:
        return (
            model1.get_criterion(criterion)
            < model0.get_criterion(criterion) - criterion_threshold
        )
    elif criterion in [
        Criterion.LH,
        Criterion.LLH,
    ]:
        return (
            model1.get_criterion(criterion)
            > model0.get_criterion(criterion) + criterion_threshold
        )
    else:
        raise NotImplementedError(f'Unknown criterion: {criterion}.')


def models_from_yaml_list(
    model_list_yaml: TYPE_PATH,
    petab_problem: petab.Problem = None,
    allow_single_model: bool = True,
) -> List[Model]:
    """Generate a model from a PEtab Select list of model YAML file.

    Args:
        model_list_yaml:
            The path to the PEtab Select list of model YAML file.
        petab_problem:
            See `Model.from_dict`.
        allow_single_model:
            Given a YAML file that contains a single model directly (not in
            a 1-element list), if `True` then the single model will be read in,
            else an error will be raised.

    Returns:
        A list of model instances, initialized with the provided
        attributes.
    """
    with open(str(model_list_yaml)) as f:
        model_dict_list = yaml.safe_load(f)
    if model_dict_list is None:
        return []

    if not isinstance(model_dict_list, list):
        if allow_single_model:
            return [
                Model.from_dict(
                    model_dict_list,
                    base_path=Path(model_list_yaml).parent,
                    petab_problem=petab_problem,
                )
            ]
        raise ValueError('The YAML file does not contain a list of models.')

    return [
        Model.from_dict(
            model_dict,
            base_path=Path(model_list_yaml).parent,
            petab_problem=petab_problem,
        )
        for model_dict in model_dict_list
    ]


def models_to_yaml_list(
    models: List[Model],
    output_yaml: TYPE_PATH,
    relative_paths: bool = True,
):
    paths_relative_to = None
    if relative_paths:
        paths_relative_to = Path(output_yaml).parent
    model_dicts = [
        model.to_dict(paths_relative_to=paths_relative_to) for model in models
    ]
    model_dicts = None if not model_dicts else model_dicts
    with open(output_yaml, 'w') as f:
        yaml.dump(model_dicts, f)



import csv
import os.path
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import petab

from .candidate_space import CandidateSpace
from .constants import (
    ESTIMATE,
    INITIAL_MODEL_METHODS,
    TYPE_PATH,
    Criterion,
    Method,
)
from .model import Model, default_compare
from .problem import Problem


def candidates(
    problem: Problem,
    candidate_space: Optional[CandidateSpace] = None,
    limit: Union[float, int] = np.inf,
    limit_sent: Union[float, int] = np.inf,
    calibrated_models: Optional[Dict[str, Model]] = None,
    newly_calibrated_models: Optional[Dict[str, Model]] = None,
    excluded_models: Optional[List[Model]] = None,
    excluded_model_hashes: Optional[List[str]] = None,
    criterion: Optional[Criterion] = None,
) -> CandidateSpace:
    """Search the model space for candidate models.

    A predecessor model is chosen from `newly_calibrated_models` if available,
    otherwise from `calibrated_models`, and is used for applicable methods.

    Args:
        problem:
            A PEtab Select problem.
        candidate_space:
            The candidate space. Defaults to a new candidate space based on the method
            defined in the problem.
        limit:
            The maximum number of models to add to the candidate space.
        limit_sent:
            The maximum number of models sent to the candidate space (which are possibly
            rejected and excluded).
        calibrated_models:
            All calibrated models in the model selection.
        newly_calibrated_models:
            All calibrated models in the most recent iteration of model
            selection.
        excluded_models:
            Models that will be excluded from model subspaces during the search for
            candidates.
        excluded_model_hashes:
            Hashes of models that will be excluded from model subspaces during the
            search for candidates.
        criterion:
            The criterion by which models will be compared. Defaults to the criterion
            defined in the PEtab Select problem.

    Returns:
        The candidate space, which contains the candidate models.
    """
    do_search = True
    # FIXME might be difficult for a CLI tool to specify a specific predecessor
    #       model if their candidate space has models. Need a way to empty
    #       the candidate space of models... might be difficult with pickled
    #       candidate space objects/arguments?
    if excluded_models is None:
        excluded_models = []
    if excluded_model_hashes is None:
        excluded_model_hashes = []
    if calibrated_models is None:
        calibrated_models = {}
    if newly_calibrated_models is None:
        newly_calibrated_models = {}
    calibrated_models.update(newly_calibrated_models)
    if criterion is None:
        criterion = problem.criterion
    if candidate_space is None:
        candidate_space = problem.new_candidate_space(limit=limit)
    candidate_space.exclude_hashes(calibrated_models)

    # Set the predecessor model to the previous predecessor model.
    # Set the new predecessor_model from the initial model or
    # by calling ui.best to find the best model to jump to if
    # this is not the first step of the search.
    predecessor_model = candidate_space.previous_predecessor_model
    if newly_calibrated_models:
        predecessor_model = problem.get_best(
            newly_calibrated_models.values(),
            criterion=criterion,
        )
        # If the new predecessor model isn't better than the previous one,
        # keep the previous one.
        # If FAMoS jumped this will not be useful, since the jumped-to model
        # can be expected to be worse than the jumped-from model, in general.
        if not default_compare(
            model0=candidate_space.previous_predecessor_model,
            model1=predecessor_model,
            criterion=criterion,
        ):
            predecessor_model = candidate_space.previous_predecessor_model

        try:
            candidate_space.update_after_calibration(
                calibrated_models=calibrated_models,
                newly_calibrated_models=newly_calibrated_models,
                criterion=criterion,
            )
        except StopIteration:
            do_search = False

        # If candidate space not Famos then ignored.
        # Else, in case we jumped to most distant in this iteration, go into
        # calibration with only the model we've jumped to.
        if (
            candidate_space.governing_method == Method.FAMOS
            and candidate_space.jumped_to_most_distant
        ):
            return candidate_space

    if (
        predecessor_model is None
        and candidate_space.method in INITIAL_MODEL_METHODS
        and calibrated_models
    ):
        predecessor_model = problem.get_best(
            models=calibrated_models.values(),
            criterion=criterion,
        )
    if predecessor_model is not None:
        candidate_space.reset(predecessor_model)

    # TODO support excluding model IDs? should be faster but may have issues, e.g.:
    #      - duplicate model IDs among multiple model subspaces
    #      - perhaps less portable if model IDs are generated differently on different
    #        computers
    problem.model_space.exclude_models(models=excluded_models)
    problem.model_space.exclude_model_hashes(
        model_hashes=excluded_model_hashes
    )
    if do_search:
        problem.model_space.search(candidate_space, limit=limit_sent)

    write_summary_tsv(
        problem=problem,
        candidate_space=candidate_space,
        previous_predecessor_model=candidate_space.previous_predecessor_model,
        predecessor_model=predecessor_model,
    )

    candidate_space.previous_predecessor_model = predecessor_model

    return candidate_space


def model_to_petab(
    model: Model,
    output_path: Optional[TYPE_PATH] = None,
) -> Dict[str, Union[petab.Problem, TYPE_PATH]]:
    """Generate the PEtab problem for a model.

    Args:
        model:
            The model.
        output_path:
            If specified, the PEtab problem will be output to files in this directory.

    Returns:
        The PEtab problem, and the path to the PEtab problem YAML file, if an output
        path is provided.
    """
    return model.to_petab(output_path=output_path)


def models_to_petab(
    models: List[Model],
    output_path_prefix: Optional[List[TYPE_PATH]] = None,
) -> List[Dict[str, Union[petab.Problem, TYPE_PATH]]]:
    """Generate the PEtab problems for a list of models.

    Args:
        models:
            The list of model.
        output_path_prefix:
            If specified, the PEtab problem will be output to files in subdirectories
            of this path, where each subdirectory corresponds to a model.

    Returns:
        The PEtab problems, and the paths to the PEtab problem YAML files, if an output
        path prefix is provided.
    """
    output_path_prefix = Path(output_path_prefix)
    result = []
    for model in models:
        output_path = output_path_prefix / model.model_id
        result.append(model_to_petab(model=model, output_path=output_path))
    return result


def best(
    problem: Problem,
    models: List[Model],
    criterion: Optional[Union[str, None]] = None,
) -> Model:
    """Get the best model from a list of models.

    Args:
        problem:
            The PEtab Select problem.
        models:
            The list of models.
        criterion:
            The criterion by which models will be compared. Defaults to
            `problem.criterion`.

    Returns:
        The best model.
    """
    # TODO return list, when multiple models are equally "best"
    return problem.get_best(models=models, criterion=criterion)


def write_summary_tsv(
    problem: Problem,
    candidate_space: Optional[CandidateSpace] = None,
    previous_predecessor_model: Optional[Union[str, Model]] = None,
    predecessor_model: Optional[Model] = None,
) -> None:
    if candidate_space.summary_tsv is None:
        return

    previous_predecessor_parameter_ids = set()
    if isinstance(previous_predecessor_model, Model):
        previous_predecessor_parameter_ids = set(
            previous_predecessor_model.get_estimated_parameter_ids_all()
        )

    if predecessor_model is None:
        predecessor_model = candidate_space.predecessor_model
    predecessor_parameter_ids = set()
    predecessor_criterion = None
    if isinstance(predecessor_model, Model):
        predecessor_parameter_ids = set(
            predecessor_model.get_estimated_parameter_ids_all()
        )
        predecessor_criterion = predecessor_model.get_criterion(
            problem.criterion
        )

    diff_parameter_ids = (
        previous_predecessor_parameter_ids.symmetric_difference(
            predecessor_parameter_ids
        )
    )

    diff_candidates_parameter_ids = []
    for candidate_model in candidate_space.models:
        candidate_parameter_ids = set(
            candidate_model.get_estimated_parameter_ids_all()
        )
        diff_candidates_parameter_ids.append(
            list(
                candidate_parameter_ids.symmetric_difference(
                    predecessor_parameter_ids
                )
            )
        )

    # FIXME remove once MostDistantCandidateSpace exists...
    method = candidate_space.method
    if (
        candidate_space.governing_method == Method.FAMOS
        and isinstance(candidate_space.predecessor_model, Model)
        and candidate_space.predecessor_model.predecessor_model_hash is None
    ):
        with open(candidate_space.summary_tsv, 'r') as f:
            if sum(1 for _ in f) > 1:
                method = Method.MOST_DISTANT

    candidate_space.write_summary_tsv(
        [
            method,
            len(candidate_space.models),
            sorted(diff_parameter_ids),
            predecessor_criterion,
            sorted(predecessor_parameter_ids),
            sorted(
                diff_candidates_parameter_ids,
                key=lambda x: [x[i] for i in range(len(x))],
            ),
        ]
    )



"""Classes and methods related to candidate spaces."""
import abc
import bisect
import copy
import csv
import logging
import os.path
import warnings
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
from more_itertools import one

from .constants import (
    ESTIMATE,
    METHOD,
    METHOD_SCHEME,
    MODELS,
    NEXT_METHOD,
    PREDECESSOR_MODEL,
    PREVIOUS_METHODS,
    TYPE_PATH,
    VIRTUAL_INITIAL_MODEL,
    VIRTUAL_INITIAL_MODEL_METHODS,
    Criterion,
    Method,
)
from .handlers import TYPE_LIMIT, LimitHandler
from .misc import snake_case_to_camel_case
from .model import Model, default_compare


class CandidateSpace(abc.ABC):
    """A base class for collecting candidate models.

    The intended use of subclasses is to identify suitable models in a model
    space, that will be provided to a model selection method for selection.

    Attributes:
        distances:
            The distances of all candidate models from the initial model.
            FIXME change list to int? Is storage of more than one value
            useful?
        predecessor_model:
            The model used for comparison, e.g. for stepwise methods.
        previous_predecessor_model:
            The previous predecessor model.
        models:
            The current set of candidate models.
        exclusions:
            A list of model hashes. Models that match a hash in `exclusions` will not
            be accepted into the candidate space. The hashes of models that are accepted
            are added to `exclusions`.
        limit:
            A handler to limit the number of accepted models.
        method:
            The model selection method of the candidate space.
        governing_method:
            Used to store the search method that governs the choice of method during
            a search. In some cases, this is always the same as the method attribute.
            An example of a difference is in the bidirectional method, where `governing_method`
            stores the bidirectional method, whereas `method` may also store the forward or
            backward methods.
        retry_model_space_search_if_no_models:
            Whether a search with a candidate space should be repeated upon failure.
            Useful for the `BidirectionalCandidateSpace`, which switches directions
            upon failure.
        summary_tsv:
            A string or `pathlib.Path`. A summary of the model selection progress
            will be written to this file.
        #limited:
        #    A descriptor that handles the limit on the number of accepted models.
        #limit:
        #    Models will fail `self.consider` if `len(self.models) >= limit`.
    """

    governing_method: Method = None
    method: Method = None
    retry_model_space_search_if_no_models: bool = False

    def __init__(
        self,
        # TODO add MODEL_TYPE = Union[str, Model], str for VIRTUAL_INITIAL_MODEL
        predecessor_model: Optional[Model] = None,
        exclusions: Optional[List[Any]] = None,
        limit: TYPE_LIMIT = np.inf,
        summary_tsv: TYPE_PATH = None,
        previous_predecessor_model: Optional[Model] = None,
    ):
        self.limit = LimitHandler(
            current=self.n_accepted,
            limit=limit,
        )
        # Each candidate class specifies this as a class attribute.
        if self.governing_method is None:
            self.governing_method = self.method
        self.reset(predecessor_model=predecessor_model, exclusions=exclusions)

        self.summary_tsv = summary_tsv
        if self.summary_tsv is not None:
            self.summary_tsv = Path(self.summary_tsv)
            self._setup_summary_tsv()

        self.previous_predecessor_model = previous_predecessor_model
        if self.previous_predecessor_model is None:
            self.previous_predecessor_model = self.predecessor_model

    def write_summary_tsv(self, row):
        if self.summary_tsv is None:
            return

        # Format single values to be valid rows
        if not isinstance(row, list):
            row = [
                row,
                *([''] * 5),
            ]

        with open(self.summary_tsv, 'a', encoding="utf-8") as f:
            writer = csv.writer(f, delimiter='\t')
            writer.writerow(row)

    def _setup_summary_tsv(self):
        self.summary_tsv.resolve().parent.mkdir(parents=True, exist_ok=True)

        if self.summary_tsv.exists():
            with open(self.summary_tsv, "r", encoding="utf-8") as f:
                last_row = f.readlines()[-1]
        else:
            self.write_summary_tsv(
                [
                    'method',
                    '# candidates',
                    'predecessor change',
                    'current model criterion',
                    'current model',
                    'candidate changes',
                ]
            )

    @classmethod
    def read_arguments_from_yaml_dict(cls, yaml_dict):
        kwargs = copy.deepcopy(yaml_dict)

        predecessor_model = None
        if (
            predecessor_model_yaml := kwargs.pop(PREDECESSOR_MODEL, None)
        ) is not None:
            predecessor_model = Model.from_yaml(predecessor_model_yaml)

        return {
            **kwargs,
            PREDECESSOR_MODEL: predecessor_model,
        }

    def is_plausible(self, model: Model) -> bool:
        """Determine whether a candidate model is plausible.

        A plausible model is one that could possibly be chosen by
        model selection method, in the absence of information about other
        models in the model space.

        For example, given a forward selection method that starts with an
        initial model `self.predecessor_model` that has no estimated
        parameters, then only models with one or more estimated parameters are
        plausible.

        Args:
            model:
                The candidate model.

        Returns:
            `True` if `model` is plausible, else `False`.
        """
        return True

    def distance(self, model: Model) -> Union[None, float, int]:
        """Compute the distance between two models that are neighbors.

        Args:
            model:
                The candidate model.
            predecessor_model:
                The initial model.

        Returns:
            The distance from `predecessor_model` to `model`, or `None` if the
            distance should not be computed.
        """
        return None

    def accept(
        self,
        model: Model,
        distance: Union[None, float, int],
        # keep_others: bool = True,
    ) -> None:
        """Add a candidate model to the candidate space.

        Args:
            model:
                The model that will be added.
            distance:
                The distance of the model from the predecessor model.
            #keep_others:
            #    Whether to keep other models that were previously added to the
            #    candidate space.
        """
        model.predecessor_model_hash = (
            self.predecessor_model.get_hash()
            if isinstance(self.predecessor_model, Model)
            else self.predecessor_model
        )
        self.models.append(model)
        self.distances.append(distance)
        self.exclude(model)

    def n_accepted(self) -> TYPE_LIMIT:
        """Get the current number of accepted models."""
        return len(self.models)

    def exclude(
        self,
        model: Union[Model, List[Model]],
    ) -> None:
        """Exclude model(s) from future consideration.

        Args:
            model:
                The model(s) that will be excluded.
        """
        if isinstance(model, list):
            for _model in model:
                self.exclusions.append(_model.get_hash())
        else:
            self.exclusions.append(model.get_hash())

    def exclude_hashes(self, hashes: List[str]) -> None:
        """Exclude models from future consideration, by hash.

        Args:
            hashes:
                The model hashes that will be excluded
        """
        self.exclusions.extend(hashes)

    def excluded(
        self,
        model: Model,
    ) -> bool:
        """Whether a model is excluded."""
        return model.get_hash() in self.exclusions

    @abc.abstractmethod
    def _consider_method(self, model) -> bool:
        """Consider whether a model should be accepted, according to a method.

        Args:
            model:
                The candidate model.

        Returns:
            Whether a model should be accepted.
        """
        return True

    def consider(self, model: Union[Model, None]) -> bool:
        """Add a candidate model, if it should be added.

        Args:
            model:
                The candidate model. This value may be `None` if the `ModelSubspace`
                decided to exclude the model that would have been sent.

        Returns:
            Whether it is OK to send additional models to the candidate space. For
            example, if the limit of the number of accepted models has been reached,
            then no further models should be sent.
            TODO change to return whether the model was accepted, and instead add
                 `self.continue` to determine whether additional models should be sent.
        """
        # Model was excluded by the `ModelSubspace` that called this method, so can be
        # skipped.
        if model is None:
            # TODO use a different code than `True`?
            return True
        if self.limit.reached():
            return False
        if self.excluded(model):
            warnings.warn(
                f'Model has been previously excluded from the candidate space so is skipped here. Model subspace ID: {model.model_subspace_id}. Parameterization: {model.parameters}',
                RuntimeWarning,
            )
            return True
        if not self.is_plausible(model):
            return True
        if not self._consider_method(model):
            return True
        self.accept(model, distance=self.distance(model))
        return not self.limit.reached()

    def reset_accepted(self) -> None:
        """Reset the accepted models."""
        self.models = []
        self.distances = []

    def set_predecessor_model(
        self, predecessor_model: Union[Model, str, None]
    ):
        self.predecessor_model = predecessor_model
        if (
            self.predecessor_model == VIRTUAL_INITIAL_MODEL
            and self.method not in VIRTUAL_INITIAL_MODEL_METHODS
        ):
            raise ValueError(
                f'A virtual initial model was requested for a method ({self.method}) that does not support them.'
            )

    def get_predecessor_model(self):
        return self.predecessor_model

    def set_exclusions(self, exclusions: Union[List[str], None]):
        # TODO change to List[str] for hashes?
        self.exclusions = exclusions
        if self.exclusions is None:
            self.exclusions = []

    def get_exclusions(self):
        return self.exclusions

    def set_limit(self, limit: TYPE_LIMIT = None):
        if limit is not None:
            self.limit.set_limit(limit)

    def get_limit(self):
        return self.limit.get_limit()

    def wrap_search_subspaces(self, search_subspaces: Callable[[], None]):
        """Decorate the subspace searches of a model space.

        Used by candidate spaces to perform changes that alter the search.
        See `BidirectionalCandidateSpace` for an example, where it's used to switch directions.

        Args:
            search_subspaces:
                The method that searches the subspaces of a model space.
        """

        def wrapper():
            search_subspaces()

        return wrapper

    def reset(
        self,
        predecessor_model: Optional[Union[Model, str, None]] = None,
        # FIXME change `Any` to some `TYPE_MODEL_HASH` (e.g. union of str/int/float)
        exclusions: Optional[Union[List[str], None]] = None,
        limit: TYPE_LIMIT = None,
    ) -> None:
        """Reset the candidate models, optionally reinitialize with a model.

        Args:
            predecessor_model:
                The initial model.
            exclusions:
                Whether to reset model exclusions.
            limit:
                The new upper limit of the number of models in this candidate space.
        """
        self.set_predecessor_model(predecessor_model)
        self.reset_accepted()
        self.set_exclusions(exclusions)
        self.set_limit(limit)

    def distances_in_estimated_parameters(
        self,
        model: Model,
        predecessor_model: Optional[Model] = None,
    ) -> Dict[str, Union[float, int]]:
        """Distance between two models in model space, using different metrics.

        All metrics are in terms of estimated parameters.

        Metrics:
            l1:
                The L_1 distance between two models.
            size:
                The difference in the number of estimated parameters between two
                models.

        Args:
            model:
                The candidate model.

        Returns:
            The distances between the models, as a dictionary, where a key is the
            name of the metric, and the value is the corresponding distance.
        """
        model0 = predecessor_model
        if model0 is None:
            model0 = self.predecessor_model
        model1 = model

        if model0 != VIRTUAL_INITIAL_MODEL and not model1.petab_yaml.samefile(
            model0.petab_yaml
        ):
            raise NotImplementedError(
                'Computation of distances between different PEtab problems is '
                'currently not supported. This error is also raised if the same '
                'PEtab problem is read from YAML files in different locations.'
            )

        # All parameters from the PEtab problem are used in the computation.
        if model0 == VIRTUAL_INITIAL_MODEL:
            parameter_ids = list(model1.petab_parameters)
            # FIXME need to take superset of all parameters amongst all PEtab problems
            # in all model subspaces to get an accurate comparable distance. Currently
            # only reasonable when working with a single PEtab problem for all models
            # in all subspaces.
            if self.method == Method.FORWARD:
                parameters0 = np.array([0 for _ in parameter_ids])
            elif self.method == Method.BACKWARD:
                parameters0 = np.array([ESTIMATE for _ in parameter_ids])
            else:
                raise NotImplementedError(
                    'Distances for the virtual initial model have not yet been '
                    f'implemented for the method "{self.method}". Please notify the'
                    'developers.'
                )
        else:
            parameter_ids = list(model0.petab_parameters)
            parameters0 = np.array(
                model0.get_parameter_values(parameter_ids=parameter_ids)
            )

        parameters1 = np.array(
            model1.get_parameter_values(parameter_ids=parameter_ids)
        )

        # TODO change to some numpy elementwise operation
        estimated0 = np.array([p == ESTIMATE for p in parameters0]).astype(int)
        estimated1 = np.array([p == ESTIMATE for p in parameters1]).astype(int)

        difference = estimated1 - estimated0

        # Changes to(/from) estimated from(/to) not estimated
        l1 = np.abs(difference).sum()

        # Change in the number of estimated parameters.
        size = np.sum(difference)

        # TODO constants?
        distances = {
            'l1': l1,
            'size': size,
        }

        return distances

    def update_after_calibration(
        self,
        *args,
        **kwargs,
    ):
        """Do work in the candidate space after calibration.

        For example, this is used by the `FamosCandidateSpace` to switch
        methods.

        Different candidate spaces require different arguments. All arguments
        are here, to ensure candidate spaces can be switched easily and still
        receive sufficient arguments.
        """
        pass


class ForwardCandidateSpace(CandidateSpace):
    """The forward method class.

    Attributes:
        direction:
            `1` for the forward method, `-1` for the backward method.
        max_steps:
            Maximum number of steps forward in a single iteration of forward selection.
            Defaults to no maximum (`None`).
    """

    method = Method.FORWARD
    direction = 1

    def __init__(
        self,
        *args,
        predecessor_model: Optional[Union[Model, str]] = None,
        max_steps: int = None,
        **kwargs,
    ):
        # Although `VIRTUAL_INITIAL_MODEL` is `str` and can be used as a default
        # argument, `None` may be passed by other packages, so the default value
        # is handled here instead.
        self.max_steps = max_steps
        if predecessor_model is None:
            predecessor_model = VIRTUAL_INITIAL_MODEL
        super().__init__(*args, predecessor_model=predecessor_model, **kwargs)

    def is_plausible(self, model: Model) -> bool:
        distances = self.distances_in_estimated_parameters(model)
        n_steps = self.direction * distances['size']

        if self.max_steps is not None and n_steps > self.max_steps:
            raise StopIteration(
                f"Maximal number of steps for method {self.method} exceeded. Stop sending candidate models."
            )

        # A model is plausible if the number of estimated parameters strictly
        # increases (or decreases, if `self.direction == -1`), and no
        # previously estimated parameters become fixed.
        if self.predecessor_model == VIRTUAL_INITIAL_MODEL or (
            n_steps > 0 and distances['l1'] == n_steps
        ):
            return True
        return False

    def distance(self, model: Model) -> int:
        # TODO calculated here and `is_plausible`. Rewrite to only calculate
        #      once?
        distances = self.distances_in_estimated_parameters(model)
        return distances['l1']

    def _consider_method(self, model) -> bool:
        """See `CandidateSpace._consider_method`."""
        distance = self.distance(model)

        # Get the distance of the current "best" plausible model(s)
        distance0 = np.inf
        if self.distances:
            distance0 = one(set(self.distances))
            # TODO store each or just one?
            # distance0 = one(self.distances)
        # breakpoint()

        # Only keep the best model(s).
        if distance > distance0:
            return False
        if distance < distance0:
            self.reset_accepted()
        return True


class BackwardCandidateSpace(ForwardCandidateSpace):
    """The backward method class."""

    method = Method.BACKWARD
    direction = -1


class BidirectionalCandidateSpace(ForwardCandidateSpace):
    """The bidirectional method class.

    Attributes:
        method_history:
            The history of models that were found at each search.
            A list of dictionaries, where each dictionary contains keys for the `METHOD`
            and the list of `MODELS`.
    """

    # TODO refactor method to inherit from governing_method if not specified
    #      by constructor argument -- remove from here.
    method = Method.BIDIRECTIONAL
    governing_method = Method.BIDIRECTIONAL
    retry_model_space_search_if_no_models = True

    def __init__(
        self,
        *args,
        initial_method: Method = Method.FORWARD,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        # FIXME cannot access from CLI
        # FIXME probably fine to replace `self.initial_method`
        #       with `self.method` here. i.e.:
        #       1. change `method` to `Method.FORWARD
        #       2. change signature to `initial_method: Method = None`
        #       3. change code here to `if initial_method is not None: self.method = initial_method`
        self.initial_method = initial_method

        self.method_history: List[Dict[str, Union[Method, List[Model]]]] = []

    def update_method(self, method: Method):
        if method == Method.FORWARD:
            self.direction = 1
        elif method == Method.BACKWARD:
            self.direction = -1
        else:
            raise NotImplementedError(
                f'Bidirectional direction must be either `Method.FORWARD` or `Method.BACKWARD`, not {method}.'
            )

        self.method = method

    def switch_method(self):
        if self.method == Method.FORWARD:
            method = Method.BACKWARD
        elif self.method == Method.BACKWARD:
            method = Method.FORWARD

        self.update_method(method=method)

    def setup_before_model_subspaces_search(self):
        # If previous search found no models, then switch method.
        previous_search = (
            None if not self.method_history else self.method_history[-1]
        )
        if previous_search is None:
            self.update_method(self.initial_method)
            return

        self.update_method(previous_search[METHOD])
        if not previous_search[MODELS]:
            self.switch_method()
            self.retry_model_space_search_if_no_models = False

    def setup_after_model_subspaces_search(self):
        current_search = {
            METHOD: self.method,
            MODELS: self.models,
        }
        self.method_history.append(current_search)
        self.method = self.governing_method

    def wrap_search_subspaces(self, search_subspaces):
        def wrapper():
            def iterate():
                self.setup_before_model_subspaces_search()
                search_subspaces()
                self.setup_after_model_subspaces_search()

            # Repeat until models are found or switching doesn't help.
            iterate()
            while (
                not self.models and self.retry_model_space_search_if_no_models
            ):
                iterate()

            # Reset flag for next time.
            self.retry_model_space_search_if_no_models = True

        return wrapper


class FamosCandidateSpace(CandidateSpace):
    """The FAMoS method class.

    This candidate space implements and extends the original FAMoS
    algorithm (doi: 10.1371/journal.pcbi.1007230).

    Attributes:
        critical_parameter_sets:
            A list of lists, where each inner list contains parameter IDs.
            All models must estimate at least 1 parameter from each critical
            parameter set.
        swap_parameter_sets:
            A list of lists, where each inner list contains parameter IDs.
            The lateral moves in FAMoS are constrained to be between parameters that
            exist in the same swap parameter set.
        method_scheme:
            A dictionary that specifies how to switch between methods when
            the current method doesn't produce a better model.
            Keys are `n`-tuples that described a pattern of length `n`
            methods. Values are methods. If the previous methods match the
            pattern in the key, then the method in the value will be used next.
            The order of the dictionary is important: only the first matched
            pattern will be used.
            Defaults to the method scheme described in the original FAMoS
            publication.
        n_reattempts:
            Integer. The total number of times that a jump-to-most-distance action
            will be performed, triggered whenever the model selection would
            normally terminate. Defaults to no reattempts (`0`).
        consecutive_laterals:
            Boolean. If `True`, the method will continue performing lateral moves
            while they produce better models. Otherwise, the method scheme will
            be applied after one lateral move.
    """

    method = Method.FAMOS
    default_method_scheme = {
        (Method.BACKWARD, Method.FORWARD): Method.LATERAL,
        (Method.FORWARD, Method.BACKWARD): Method.LATERAL,
        (Method.BACKWARD, Method.LATERAL): None,
        (Method.FORWARD, Method.LATERAL): None,
        (Method.FORWARD,): Method.BACKWARD,
        (Method.BACKWARD,): Method.FORWARD,
        (Method.LATERAL,): Method.FORWARD,
        (Method.MOST_DISTANT,): Method.FORWARD,
        None: Method.FORWARD,
    }

    def __init__(
        self,
        *args,
        predecessor_model: Optional[Union[Model, str, None]] = None,
        critical_parameter_sets: List = [],
        swap_parameter_sets: List = [],
        method_scheme: Dict[tuple, str] = None,
        n_reattempts: int = 0,
        consecutive_laterals: bool = False,
        **kwargs,
    ):
        self.critical_parameter_sets = critical_parameter_sets
        self.swap_parameter_sets = swap_parameter_sets

        self.method_scheme = method_scheme
        if method_scheme is None:
            self.method_scheme = self.default_method_scheme

        # FIXME remove and use `self.method` everywhere in the constructor
        #       instead -- not required in other methods anymore
        self.initial_method = self.method_scheme[None]
        self.method = self.initial_method
        self.method_history = [self.initial_method]

        if predecessor_model is None:
            predecessor_model = VIRTUAL_INITIAL_MODEL

        if (
            predecessor_model == VIRTUAL_INITIAL_MODEL
            and critical_parameter_sets
        ) or (
            predecessor_model != VIRTUAL_INITIAL_MODEL
            and not self.check_critical(predecessor_model)
        ):
            raise ValueError(
                f'Provided predecessor model {predecessor_model.parameters} does not contain necessary critical parameters {self.critical_parameter_sets}. Provide a valid predecessor model.'
            )

        if (
            predecessor_model == VIRTUAL_INITIAL_MODEL
            and self.initial_method not in VIRTUAL_INITIAL_MODEL_METHODS
        ):
            raise ValueError(
                f"The initial method {self.initial_method} does not support the `VIRTUAL_INITIAL_MODEL` as its predecessor_model."
            )

        # FIXME remove `None` from the resulting `inner_methods` set?
        inner_methods = set.union(
            *[
                set(
                    [
                        *(
                            method_pattern
                            if method_pattern is not None
                            else (None,)
                        ),
                        next_method,
                    ]
                )
                for method_pattern, next_method in self.method_scheme.items()
            ]
        )
        if Method.LATERAL in inner_methods and not self.swap_parameter_sets:
            raise ValueError(
                f"Use of the lateral method with FAMoS requires `swap_parameter_sets`."
            )

        for method in inner_methods:
            if method is not None and method not in [
                Method.FORWARD,
                Method.LATERAL,
                Method.BACKWARD,
                Method.MOST_DISTANT,
            ]:
                raise NotImplementedError(
                    f'Methods FAMoS can swap to are `Method.FORWARD`, `Method.BACKWARD` and `Method.LATERAL`, not {method}. \
                    Check if the method_scheme scheme provided is correct.'
                )

        self.inner_candidate_spaces = {
            Method.FORWARD: ForwardCandidateSpace(
                *args,
                predecessor_model=predecessor_model,
                **kwargs,
            ),
            Method.BACKWARD: BackwardCandidateSpace(
                *args,
                predecessor_model=predecessor_model,
                **kwargs,
            ),
            Method.LATERAL: LateralCandidateSpace(
                *args,
                predecessor_model=(
                    predecessor_model
                    if predecessor_model != VIRTUAL_INITIAL_MODEL
                    else None
                ),
                max_steps=1,
                **kwargs,
            ),
        }
        self.inner_candidate_space = self.inner_candidate_spaces[
            self.initial_method
        ]

        super().__init__(*args, predecessor_model=predecessor_model, **kwargs)

        self.governing_method = Method.FAMOS

        self.n_reattempts = n_reattempts

        self.consecutive_laterals = consecutive_laterals
        if (
            not self.consecutive_laterals
            and (Method.LATERAL,) not in self.method_scheme
        ):
            raise ValueError(
                "Please provide a method to switch to after a lateral search, if not enabling the `consecutive_laterals` option."
            )

        if self.n_reattempts:
            # TODO make so max_number can be specified? It cannot in original FAMoS.
            self.most_distant_max_number = 100
        else:
            self.most_distant_max_number = 1

        self.best_models = []
        self.best_model_of_current_run = predecessor_model

        self.jumped_to_most_distant = False
        self.swap_done_successfully = False

    @classmethod
    def read_arguments_from_yaml_dict(cls, yaml_dict) -> dict:
        kwargs = copy.deepcopy(yaml_dict)

        if (method_scheme_raw := kwargs.pop(METHOD_SCHEME, None)) is not None:
            method_scheme = {
                (
                    tuple(
                        [
                            Method(method_str)
                            for method_str in definition[PREVIOUS_METHODS]
                        ]
                    )
                    if definition[PREVIOUS_METHODS] is not None
                    else None
                ): definition[NEXT_METHOD]
                for definition in method_scheme_raw
            }
            kwargs[METHOD_SCHEME] = method_scheme

        return super().read_arguments_from_yaml_dict(yaml_dict=kwargs)

    def update_after_calibration(
        self,
        *args,
        calibrated_models: Dict[str, Model],
        newly_calibrated_models: Dict[str, Model],
        criterion: Criterion,
        **kwargs,
    ) -> None:
        """See `CandidateSpace.update_after_calibration`."""

        # super().update_after_calibration(*args, **kwargs)

        # In case we jumped to most distant in the last iteration,
        # there's no need for an update, so we reset the jumped variable
        # to False and continue to candidate generation
        if self.jumped_to_most_distant:
            self.jumped_to_most_distant = False
            jumped_to_model = one(newly_calibrated_models.values())
            self.set_predecessor_model(jumped_to_model)
            self.best_model_of_current_run = jumped_to_model
            return False

        if self.update_from_newly_calibrated_models(
            newly_calibrated_models=newly_calibrated_models,
            criterion=criterion,
        ):
            logging.info("Switching method")
            self.switch_method(calibrated_models=calibrated_models)
            self.switch_inner_candidate_space(
                calibrated_models=calibrated_models,
            )
            logging.info(
                "Method switched to ", self.inner_candidate_space.method
            )

        self.method_history.append(self.method)

    def update_from_newly_calibrated_models(
        self,
        newly_calibrated_models: Dict[str, Model],
        criterion: Criterion,
    ) -> bool:
        """Update the self.best_models with the latest
        `newly_calibrated_models`
        and determine if there was a new best model. If so, return
        True. False otherwise."""

        go_into_switch_method = True
        for newly_calibrated_model in newly_calibrated_models.values():
            if (
                self.best_model_of_current_run == VIRTUAL_INITIAL_MODEL
                or default_compare(
                    model0=self.best_model_of_current_run,
                    model1=newly_calibrated_model,
                    criterion=criterion,
                )
            ):
                go_into_switch_method = False
                self.best_model_of_current_run = newly_calibrated_model

            if len(
                self.best_models
            ) < self.most_distant_max_number or default_compare(
                model0=self.best_models[self.most_distant_max_number - 1],
                model1=newly_calibrated_model,
                criterion=criterion,
            ):
                self.insert_model_into_best_models(
                    model_to_insert=newly_calibrated_model,
                    criterion=criterion,
                )

        self.best_models = self.best_models[: self.most_distant_max_number]

        # When we switch to LATERAL method, we will do only one iteration with this
        # method. So if we do it succesfully (i.e. that we found a new best model), we
        # want to switch method. This is why we put go_into_switch_method to True, so
        # we go into the method switching pipeline
        if self.method == Method.LATERAL and not self.consecutive_laterals:
            self.swap_done_successfully = True
            go_into_switch_method = True
        return go_into_switch_method

    def insert_model_into_best_models(
        self, model_to_insert: Model, criterion: Criterion
    ) -> None:
        """Inserts a model into the list of best_models which are sorted
        w.r.t. the criterion specified."""
        insert_index = bisect.bisect_left(
            [model.get_criterion(criterion) for model in self.best_models],
            model_to_insert.get_criterion(criterion),
        )
        self.best_models.insert(insert_index, model_to_insert)

    def consider(self, model: Union[Model, None]) -> bool:
        """Re-define consider of FAMoS to be the consider method
        of the inner_candidate_space. Update all the attributes
        changed in the condsider method."""

        if self.limit.reached():
            return False

        # Check if model contains necessary critical parameters and, if
        # the current method is swap, check the swap move is contained
        # in a swap set.

        if self.check_critical(model) and (
            not self.method == Method.LATERAL or self.check_swap(model)
        ):
            return_value = self.inner_candidate_space.consider(model)

            # update the attributes
            self.models = self.inner_candidate_space.models
            self.distances = self.inner_candidate_space.distances
            self.exclusions = self.inner_candidate_space.exclusions

            return return_value

        return True

    def _consider_method(self, model) -> bool:
        """See `CandidateSpace._consider_method`."""
        return self.inner_candidate_space._consider_method(model)

    def reset_accepted(self) -> None:
        """Changing the reset_accepted to reset the
        inner_candidate_space as well."""
        super().reset_accepted()
        self.inner_candidate_space.reset_accepted()

    def set_predecessor_model(
        self, predecessor_model: Union[Model, str, None]
    ):
        """Setting the predecessor model for the
        inner_candidate_space as well."""
        super().set_predecessor_model(predecessor_model=predecessor_model)
        self.inner_candidate_space.set_predecessor_model(
            predecessor_model=predecessor_model
        )

    def set_exclusions(self, exclusions: Union[List[str], None]):
        """Setting the exclusions for the
        inner_candidate_space as well."""
        self.exclusions = exclusions
        self.inner_candidate_space.exclusions = exclusions
        if self.exclusions is None:
            self.exclusions = []
            self.inner_candidate_space.exclusions = []

    def set_limit(self, limit: TYPE_LIMIT = None):
        """Setting the limit for the
        inner_candidate_space as well."""
        if limit is not None:
            self.limit.set_limit(limit)
            self.inner_candidate_space.limit.set_limit(limit)

    def is_plausible(self, model: Model) -> bool:
        if self.check_critical(model):
            return self.inner_candidate_space.is_plausible(model)
        return False

    def check_swap(self, model: Model) -> bool:
        """Check if parameters that are swapped are contained in the
        same swap parameter set."""
        if self.method != Method.LATERAL:
            return True

        predecessor_estimated_parameters_ids = set(
            self.predecessor_model.get_estimated_parameter_ids_all()
        )
        estimated_parameters_ids = set(model.get_estimated_parameter_ids_all())

        swapped_parameters_ids = estimated_parameters_ids.symmetric_difference(
            predecessor_estimated_parameters_ids
        )

        for swap_set in self.swap_parameter_sets:
            if swapped_parameters_ids.issubset(set(swap_set)):
                return True
        return False

    def check_critical(self, model: Model) -> bool:
        """Check if the model contains all necessary critical parameters"""

        estimated_parameters_ids = set(model.get_estimated_parameter_ids_all())
        for critical_set in self.critical_parameter_sets:
            if not estimated_parameters_ids.intersection(set(critical_set)):
                return False
        return True

    def switch_method(
        self,
        calibrated_models: Dict[str, Model],
    ) -> None:
        """Switch to the next method with respect to the history
        of methods used and the switching scheme in self.method_scheme"""

        previous_method = self.method
        next_method = previous_method
        logging.info("SWITCHING", self.method_history)
        # If last method was LATERAL and we have made a succesfull swap
        # (a better model was found) then go back to FORWARD. Else do the
        # usual method swapping scheme.
        if self.swap_done_successfully:
            next_method = self.method_scheme[(Method.LATERAL,)]
        else:
            # iterate through the method_scheme dictionary to see which method to switch to
            for previous_methods in self.method_scheme:
                if previous_methods is not None and previous_methods == tuple(
                    self.method_history[-len(previous_methods) :]
                ):
                    next_method = self.method_scheme[previous_methods]
                    # if found a switch just break (choosing first good switch)
                    break

        # raise error if the method didn't change
        if next_method == previous_method:
            raise ValueError(
                "Method didn't switch when it had to. "
                "The `method_scheme` provided is not sufficient. "
                "Please provide a correct method_scheme scheme. "
                f"Method history: `{self.method_history}`. "
            )

        # Terminate if next method is `None`
        if next_method is None:
            if self.n_reattempts:
                self.jump_to_most_distant(calibrated_models=calibrated_models)
                return
            raise StopIteration(
                f"The next method is {next_method}. The search is terminating."
            )

        # If we try to switch to SWAP method but it's not available (no crit or swap parameter groups)
        if (
            next_method == Method.LATERAL
            and not [
                critical_set
                for critical_set in self.critical_parameter_sets
                if len(critical_set) > 1
            ]
            and not self.swap_parameter_sets
        ):
            if self.n_reattempts:
                self.jump_to_most_distant(calibrated_models=calibrated_models)
                return
            raise ValueError(
                f"The next method is {next_method}, but there are no critical or swap parameters sets. Terminating."
            )
        if previous_method == Method.LATERAL:
            self.swap_done_successfully = False
        self.update_method(method=next_method)

    def update_method(self, method: Method):
        """Update self.method to the method."""

        self.method = method

    def switch_inner_candidate_space(
        self,
        calibrated_models: Dict[str, Model],
    ):
        """Switch self.inner_candidate_space to the candidate space of
        the current self.method."""

        # if self.method != Method.MOST_DISTANT:
        self.inner_candidate_space = self.inner_candidate_spaces[self.method]
        # reset the next inner candidate space with the current history of all
        # calibrated models
        self.inner_candidate_space.reset(
            predecessor_model=self.predecessor_model,
            exclusions=calibrated_models,
        )

    def jump_to_most_distant(
        self,
        calibrated_models: Dict[str, Model],
    ):
        """Jump to most distant model with respect to the history of all
        calibrated models."""

        predecessor_model = self.get_most_distant(
            calibrated_models=calibrated_models,
        )

        logging.info("JUMPING: ", predecessor_model.parameters)

        # if model not appropriate make it so by adding the first
        # critical parameter from each critical parameter set
        if not self.check_critical(predecessor_model):
            for critical_set in self.critical_parameter_sets:
                predecessor_model.parameters[critical_set[0]] = ESTIMATE

        # self.update_method(self.initial_method)
        self.update_method(Method.MOST_DISTANT)

        self.n_reattempts -= 1
        self.jumped_to_most_distant = True

        # FIXME rename here `predecessor_model` to `most_distant_model`
        # self.predecessor_model = None
        self.set_predecessor_model(None)
        self.best_model_of_current_run = None
        self.models = [predecessor_model]

        self.write_summary_tsv("Jumped to the most distant model.")
        self.update_method(self.method_scheme[(Method.MOST_DISTANT,)])

    # TODO Fix for non-famos model subspaces. FAMOS easy because of only 0;ESTIMATE
    def get_most_distant(
        self,
        calibrated_models: Dict[str, Model],
    ) -> Model:
        """
        Get most distant model to all the checked models. We take models from the
        sorted list of best models (self.best_models) and construct complements of
        these models. For all these complements we compute the distance in number of
        different estimated parameters to all models from history. For each complement
        we take the minimum of these distances as it's distance to history. Then we
        choose the complement model with the maximal distance to history.
        TODO:
        Next we check if this model is contained in any subspace. If so we choose it.
        If not we choose the model in a subspace that has least distance to this
        complement model.
        """

        most_distance = 0
        most_distant_indices = []

        parameter_ids = self.best_models[0].petab_parameters

        for model in self.best_models:
            model_estimated_parameters = np.array(
                [
                    p == ESTIMATE
                    for p in model.get_parameter_values(
                        parameter_ids=parameter_ids
                    )
                ]
            ).astype(int)
            complement_parameters = 1 - model_estimated_parameters
            # initialize the least distance to the maximal possible value of it
            complement_least_distance = len(complement_parameters)
            # get the complement least distance
            for calibrated_model in calibrated_models.values():
                calibrated_model_estimated_parameters = np.array(
                    [
                        p == ESTIMATE
                        for p in calibrated_model.get_parameter_values(
                            parameter_ids=parameter_ids
                        )
                    ]
                ).astype(int)

                difference = (
                    calibrated_model_estimated_parameters
                    - complement_parameters
                )
                l1_distance = np.abs(difference).sum()

                if l1_distance < complement_least_distance:
                    complement_least_distance = l1_distance

            # if the complement is further away than the current best one
            # then save it as the new most distant away one
            if complement_least_distance > most_distance:
                most_distance = complement_least_distance
                most_distant_indices = complement_parameters
        if len(most_distant_indices) == 0:
            raise StopIteration("No most_distant model found. Terminating")

        most_distant_parameter_values = [
            str(index).replace('1', ESTIMATE) for index in most_distant_indices
        ]

        most_distant_parameters = {
            parameter_id: index
            for parameter_id, index in zip(
                parameter_ids, most_distant_parameter_values
            )
        }

        most_distant_model = Model(
            petab_yaml=model.petab_yaml,
            model_subspace_id=model.model_subspace_id,
            model_subspace_indices=most_distant_indices,
            parameters=most_distant_parameters,
        )

        return most_distant_model

    def wrap_search_subspaces(self, search_subspaces):
        def wrapper():
            search_subspaces(only_one_subspace=True)

        return wrapper


# TODO rewrite so BidirectionalCandidateSpace inherits from ForwardAndBackwardCandidateSpace
#      instead
class ForwardAndBackwardCandidateSpace(BidirectionalCandidateSpace):
    method = Method.FORWARD_AND_BACKWARD
    governing_method = Method.FORWARD_AND_BACKWARD
    retry_model_space_search_if_no_models = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, initial_method=None)

    def wrap_search_subspaces(self, search_subspaces):
        def wrapper():
            for method in [Method.FORWARD, Method.BACKWARD]:
                self.update_method(method=method)
                search_subspaces()
                self.setup_after_model_subspaces_search()

        return wrapper

    # Disable unused interface
    setup_before_model_subspaces_search = None
    switch_method = None

    def setup_after_model_space_search(self):
        pass


class LateralCandidateSpace(CandidateSpace):
    """Find models with the same number of estimated parameters."""

    method = Method.LATERAL

    def __init__(
        self,
        *args,
        predecessor_model: Union[Model, None],
        max_steps: int = None,
        **kwargs,
    ):
        """
        Additional args:
            max_number_of_steps:
                Maximal allowed number of swap moves. If 0 then there is no maximum.
        """
        super().__init__(
            *args,
            predecessor_model=predecessor_model,
            **kwargs,
        )
        self.max_steps = max_steps

    def is_plausible(self, model: Model) -> bool:
        if self.predecessor_model is None:
            raise ValueError(
                f"The predecessor_model is still None. Provide an appropriate predecessor_model"
            )

        distances = self.distances_in_estimated_parameters(model)

        # If max_number_of_steps is non-zero and the number of steps made is
        # larger then move is not plausible.
        if self.max_steps is not None and distances['l1'] > 2 * self.max_steps:
            raise StopIteration(
                f"Maximal number of steps for method {self.method} exceeded. Stop sending candidate models."
            )

        # A model is plausible if the number of estimated parameters remains
        # the same, but some estimated parameters have become fixed and vice
        # versa.
        if (
            distances['size'] == 0
            and
            # distances['size'] == 0 implies L1 % 2 == 0.
            # FIXME here and elsewhere, deal with models that are equal
            #       except for the values of their fixed parameters.
            distances['l1'] > 0
        ):
            return True
        return False

    def _consider_method(self, model):
        return True


class BruteForceCandidateSpace(CandidateSpace):
    """The brute-force method class."""

    method = Method.BRUTE_FORCE

    def __init__(self, *args, **kwargs):
        # if args or kwargs:
        #    # FIXME remove?
        #    # FIXME at least support limit
        #    warnings.warn(
        #        'Arguments were provided but will be ignored, because of the '
        #        'brute force candidate space.'
        #    )
        super().__init__(*args, **kwargs)

    def _consider_method(self, model):
        return True


candidate_space_classes = [
    ForwardCandidateSpace,
    BackwardCandidateSpace,
    BidirectionalCandidateSpace,
    LateralCandidateSpace,
    BruteForceCandidateSpace,
    ForwardAndBackwardCandidateSpace,
    FamosCandidateSpace,
]


def method_to_candidate_space_class(method: Method) -> str:
    """Instantiate a candidate space given its method name.

    Args:
        method:
            The name of the method corresponding to one of the implemented candidate
            spaces.

    Returns:
        The candidate space.
    """
    for candidate_space_class in candidate_space_classes:
        if candidate_space_class.method == method:
            return candidate_space_class
    raise NotImplementedError(
        f'The provided method name {method} does not correspond to an implemented candidate space.'
    )


'''
def distances_in_estimated_parameters(
    model: Model,
    predecessor_model: Model,
    #method: str = None,
) -> Dict[str, Union[float, int]]:
    """Distance between two models in model space, using different metrics.

    All metrics are in terms of estimated parameters.

    Metrics:
        l1:
            The L_1 distance between two models.
        size:
            The difference in the number of estimated parameters between two
            models.

    Args:
        model:
            The candidate model.
        predecessor_model:
            The initial model.
        method:
            The search method in use. Necessary if the predecessor model is the
            virtual initial model.

    Returns:
        The distances between the models, as a dictionary, where a key is the
        name of the metric, and the value is the corresponding distance.
    """
    model0 = predecessor_model
    model1 = model

    if not model1.petab_yaml.samefile(model0.petab_yaml):
        raise NotImplementedError(
            'Computation of distances between different PEtab problems is '
            'currently not supported. This error is also raised if the same '
            'PEtab problem is read from YAML files in different locations.'
        )

    # All parameters from the PEtab problem are used in the computation.
    parameter_ids = list(model0.petab_parameters)

    parameters0 = np.array(model0.get_parameter_values(parameter_ids=parameter_ids))
    parameters1 = np.array(model1.get_parameter_values(parameter_ids=parameter_ids))

    estimated0 = (parameters0 == ESTIMATE).astype(int)
    estimated1 = (parameters1 == ESTIMATE).astype(int)

    difference = estimated1 - estimated0

    # Changes to(/from) estimated from(/to) not estimated
    l1 = np.abs(difference).sum()

    # Change in the number of estimated parameters.
    size = np.sum(difference)

    # TODO constants?
    distances = {
        'l1': l1,
        'size': size,
    }

    return distances
'''



import os
import re
import sys

from setuptools import find_packages, setup

org = 'PEtab-dev'
repo = 'petab_select'


def read(fname):
    """Read a file."""
    return open(fname).read()


def absolute_links(txt):
    """Replace relative petab github links by absolute links."""

    raw_base = f"(https://raw.githubusercontent.com/{org}/{repo}/main/"
    embedded_base = f"(https://github.com/{org}/{repo}/tree/main/"
    # iterate over links
    for var in re.findall(r'\[.*?\]\((?!http).*?\)', txt):
        if re.match(r'.*?.(png|svg)\)', var):
            # link to raw file
            rep = var.replace("(", raw_base)
        else:
            # link to github embedded file
            rep = var.replace("(", embedded_base)
        txt = txt.replace(var, rep)
    return txt


# 3.7.1 for NumPy, 3.8 for `typing.get_args`
minimum_python_version = '3.8'
if sys.version_info < tuple(map(int, minimum_python_version.split('.'))):
    sys.exit(f'PEtab Select requires Python >= {minimum_python_version}')

# read version from file
__version__ = ''
version_file = os.path.join('petab_select', 'version.py')
# sets __version__
exec(read(version_file))  # pylint: disable=W0122 # nosec

ENTRY_POINTS = {
    'console_scripts': [
        'petab_select = petab_select.cli:cli',
    ]
}

# project metadata
# noinspection PyUnresolvedReferences
setup(
    name='petab_select',
    version=__version__,
    description='PEtab Select: an extension to PEtab for model selection.',
    long_description=absolute_links(read('README.md')),
    long_description_content_type="text/markdown",
    # author='The PEtab Select developers',
    # author_email='dilan.pathirana@uni-bonn.de',
    url=f'https://github.com/{org}/{repo}',
    packages=find_packages(exclude=['doc*', 'test*']),
    install_requires=[
        # TODO minimum versions
        'more-itertools',
        'numpy',
        'pandas',
        'petab',
        'pyyaml',
        #'python-libsbml>=5.17.0',
        #'sympy',
        # required for CLI
        'click',
        'dill',
        # plotting
        #'matplotlib>=2.2.3',
        #'seaborn',
    ],
    include_package_data=True,
    python_requires=f'>={minimum_python_version}',
    entry_points=ENTRY_POINTS,
    extras_require={
        'test': [
            'pytest >= 5.4.3',
            'pytest-cov >= 2.10.0',
            'amici >= 0.11.25',
            'fides >= 0.7.5',
            # FIXME
            'pypesto > 0.2.13',
            # 'pypesto @ git+https://github.com/ICB-DCM/pyPESTO.git@develop#egg=pypesto',
        ],
        #'reports': ['Jinja2'],
        #'combine': ['python-libcombine>=0.2.6'],
        #'doc': [
        #    'sphinx>=3.5.3',
        #    'sphinxcontrib-napoleon>=0.7',
        #    'sphinx-markdown-tables>=0.0.15',
        #    'sphinx-rtd-theme>=0.5.1',
        #    'recommonmark>=0.7.1',
        #    'nbsphinx>=0.8.2',
        #    'm2r>=0.2.1',
        #    'ipython>=7.21.0',
        # ]
    },
)
